//
// Created by philipp on 30.11.17.
//

#ifndef PROTOCOL_VEHICLE_H
#define PROTOCOL_VEHICLE_H

#include <thread>
#include <NetworkingLib/DatagramReceiver.h>
#include <NetworkingLib/DatagramSender.h>
#include "PlatoonService.h"
#include "NetworkingLib/Networking.h"
#include "NetworkingLib/ServiceServer.h"
#include "NetworkingLib/ServiceClient.h"
#include "NetworkingLib/Resolver.h"
#include "NetworkInfo.h"

namespace platoonProtocol
{
class Vehicle : public std::enable_shared_from_this<Vehicle>
{
public:
    using Ptr = std::shared_ptr<Vehicle>;

    using Callback = std::function<void()>;

    static constexpr std::uint16_t UDP_PORT = 10000;

    enum class Role
    {
        LEADER, FOLLOWER
    };

    virtual ~Vehicle()
    {}

    bool isIdle();

    bool isPlatoonRunning();

    void createPlatoon();

    void leavePlatoon();

    virtual void stop();

    VehicleId getOwnVehicleId() const
    { return myInfo.vehicleId; }

    Role getRole() const noexcept
    { return role; }

    void setOnRunningPlatoonCallback(const Callback & callback)
    { onRunningPlatoonCallback = callback; }

    void setOnLeavingPlatoonCallback(const Callback & callback)
    { onLeavingPlatoonCallback = callback; }

protected:
    Vehicle(
        networking::Networking & net,
        const NetworkInfo & info,
        Role role);

    enum class State
    {
        IDLE,
        CREATING_PLATOON,
        RUNNING_PLATOON
    };

    static const PlatoonConfig DEFAULT_PLATOON_CONFIG;
    static const Callback DEFAULT_CALLBACK;

    const NetworkInfo myInfo;
    PlatoonId platoonId;
    std::atomic<PlatoonConfig> platoonConfig{DEFAULT_PLATOON_CONFIG};
    State state{State::IDLE};
    std::mutex stateMutex;
    networking::Networking & net;
    networking::message::DatagramReceiver<PlatoonMessage>::Ptr receiver;
    networking::message::DatagramSender<PlatoonMessage>::Ptr sender;
    Callback onRunningPlatoonCallback{DEFAULT_CALLBACK};
    Callback onLeavingPlatoonCallback{DEFAULT_CALLBACK};

    virtual void doCreatePlatoon();

    virtual void doRunPlatoon();

    virtual void doLeavePlatoon();

    virtual void receiveMessage(const PlatoonMessage & message) = 0;

    void sendMessage(const PlatoonMessage & message, const networking::time::Duration & timeout);

    void setState(State newState);

    void receive();

private:
    Role role;

};
}


#endif //PROTOCOL_VEHICLE_H
